# ============================================================================== #
# MIT License                                                                    #
#                                                                                #
# Copyright (c) 2024 Nathan Juraj Michlo                                         #
#                                                                                #
# Permission is hereby granted, free of charge, to any person obtaining a copy   #
# of this software and associated documentation files (the "Software"), to deal  #
# in the Software without restriction, including without limitation the rights   #
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell      #
# copies of the Software, and to permit persons to whom the Software is          #
# furnished to do so, subject to the following conditions:                       #
#                                                                                #
# The above copyright notice and this permission notice shall be included in all #
# copies or substantial portions of the Software.                                #
#                                                                                #
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR     #
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,       #
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE    #
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER         #
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,  #
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE  #
# SOFTWARE.                                                                      #
# ============================================================================== #


import argparse
import logging
import typing

from pydependence._cli import pydeps
from pydependence._core.requirements_map import NoConfiguredRequirementMappingError

LOGGER = logging.getLogger(__name__)

# ========================================================================= #
# CLI                                                                       #
# ========================================================================= #


if typing.TYPE_CHECKING:

    class PyDepsCliArgsProto(typing.Protocol):
        config: str


def _parse_args() -> "PyDepsCliArgsProto":
    """
    Make argument parser, that has a `--file` argument which is required, then
    returns the parsed arguments.
    """
    parser = argparse.ArgumentParser(
        description="PyDependence: A tool for scanning and resolving python dependencies across files."
    )
    parser.add_argument(
        "config",
        type=str,
        help="The python file to analyse for dependencies.",
    )
    return parser.parse_args()


def _cli():
    # args
    args = _parse_args()

    # run
    try:
        pydeps(config_path=args.config)
    except NoConfiguredRequirementMappingError as e:
        LOGGER.critical(
            f"[pydependence] no configured requirement mapping found, either specify all missing version mappings or disable strict mode:\n{e}"
        )
        exit(1)


if __name__ == "__main__":
    # set default log level to info
    logging.basicConfig(level=logging.INFO)
    # run cli
    _cli()


# ========================================================================= #
# END                                                                       #
# ========================================================================= #



# ============================================================================== #
# MIT License                                                                    #
#                                                                                #
# Copyright (c) 2024 Nathan Juraj Michlo                                         #
#                                                                                #
# Permission is hereby granted, free of charge, to any person obtaining a copy   #
# of this software and associated documentation files (the "Software"), to deal  #
# in the Software without restriction, including without limitation the rights   #
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell      #
# copies of the Software, and to permit persons to whom the Software is          #
# furnished to do so, subject to the following conditions:                       #
#                                                                                #
# The above copyright notice and this permission notice shall be included in all #
# copies or substantial portions of the Software.                                #
#                                                                                #
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR     #
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,       #
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE    #
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER         #
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,  #
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE  #
# SOFTWARE.                                                                      #
# ============================================================================== #

import logging
import warnings
from collections import defaultdict
from enum import Enum
from pathlib import Path
from typing import Dict, List, Literal, Optional, Union

import pydantic
from packaging.requirements import Requirement
from typing_extensions import Annotated

from pydependence._core.module_imports_ast import ManualImportInfo
from pydependence._core.modules_scope import (
    ModulesScope,
    RestrictMode,
    RestrictOp,
    UnreachableModeEnum,
)
from pydependence._core.requirements_map import (
    DEFAULT_REQUIREMENTS_ENV,
    ImportMatcherBase,
    ImportMatcherGlobs,
    ImportMatcherScope,
    NoConfiguredRequirementMappingError,
    ReqMatcher,
    RequirementsMapper,
)
from pydependence._core.requirements_out import OutMappedRequirements
from pydependence._core.utils import (
    apply_root_to_path_str,
    load_toml_document,
    toml_file_replace_array,
    txt_file_dump,
)

# python 3.8 support
# TODO: this pattern is not yet supported by the lazy dependency resolver
#       we should specifically add support for this pattern, as it is a common.
# TODO: we should add the ability to exclude certain imports from the pydependence
#       resolve to handle cases like this.
# try:
#     from typing import Annotated
# except ImportError:
#     from typing_extensions import Annotated

LOGGER = logging.getLogger(__name__)

# ========================================================================= #
# CONFIGS                                                                   #
# ========================================================================= #


class _ResolveRules(pydantic.BaseModel, extra="forbid"):

    # If true, then vist all the lazy imports. Usually the lazy imports are removed from
    # the import graph and we don't traverse these edges. This on the other-hand allows
    # all these edges to be traversed. This is often useful if you want to create
    # packages that only require some minimal set of requirements, and everything else
    # that you define should be optional. Also useful if you want to generate a minimal
    # dependencies list, and then in optional dependency lists you want to create a full
    # set of requirements for everything!
    visit_lazy: Optional[bool] = None

    # only applicable when `visit_lazy=False`, then in this case we re-add the lazy
    # imports that are directly referenced in all the traversed files, i.e. it is a
    # shallow include of lazy imports without continuing to traverse them. This means
    # that we add the lazy imports but we don't continue traversing. This is often not
    # useful with complex lazy import graphs that continue to reference more module
    # within the same scope as this could cause missing imports, rather specify
    # `visit_lazy=True` in this case.
    # * [A.K.A.] `shallow_include_lazy=True`
    re_add_lazy: Optional[bool] = None

    # If true, then exclude imports that were not encountered as we traversed the import
    # graph. [NOTE]: this is probably useful if you don't want to include all imports
    # below a specific scope, but only want to resolve what is actually encountered.
    # Not entirely sure this has much of an effect?
    exclude_unvisited: Optional[bool] = None

    # If true, then exclude all imports that are part of the current scope. This usually
    # should not have any effect because imports are replaced as we traverse the graph
    # through the current scope, [NOTE] thus not entirely sure that this has any effect,
    # should it be a bug if we encounter any of these?
    exclude_in_search_space: Optional[bool] = None

    # If true, then exclude all the python builtin package names from being output in
    # the requirements files. This usually should be true unless you are trying to debug
    # as this would generate invalid requirements list as these would not exist on pypi.
    exclude_builtins: Optional[bool] = None

    # Check that generated imports and requirements have entries in the versions list.
    # If strict mode is enabled, then an error is thrown if a version entry is missing.
    # If strict mode is disabled, then a warning should be given, and the root import
    # name is used instead of the requirement name, which may or may not match up
    # to an actual python package.
    strict_requirements_map: Optional[bool] = None

    # TODO: we should add some sort of option to ensure that generated dependency lists
    #       exactly match some pre-defined set, while also outputting this set.

    # TODO: we do not yet ensure that raw added requirements do not conflict with
    #       generated requirements.

    @classmethod
    def make_default_base_rules(cls):
        return _ResolveRules(
            visit_lazy=False,
            re_add_lazy=False,
            exclude_unvisited=True,
            exclude_in_search_space=True,
            exclude_builtins=True,
            strict_requirements_map=True,
        )

    def set_defaults(self, defaults: "_ResolveRules"):
        assert defaults.visit_lazy is not None
        assert defaults.re_add_lazy is not None
        assert defaults.exclude_unvisited is not None
        assert defaults.exclude_in_search_space is not None
        assert defaults.exclude_builtins is not None
        assert defaults.strict_requirements_map is not None
        if self.visit_lazy is None:
            self.visit_lazy = defaults.visit_lazy
        if self.re_add_lazy is None:
            self.re_add_lazy = defaults.re_add_lazy
        if self.exclude_unvisited is None:
            self.exclude_unvisited = defaults.exclude_unvisited
        if self.exclude_in_search_space is None:
            self.exclude_in_search_space = defaults.exclude_in_search_space
        if self.exclude_builtins is None:
            self.exclude_builtins = defaults.exclude_builtins
        if self.strict_requirements_map is None:
            self.strict_requirements_map = defaults.strict_requirements_map


# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #
# CONFIG - OUTPUT                                                           #
# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #


class OutputModeEnum(str, Enum):
    requirements = "requirements"
    optional_dependencies = "optional-dependencies"
    dependencies = "dependencies"


class _Output(_ResolveRules, extra="forbid"):
    # resolve
    scope: Optional[str] = None
    start_scope: Optional[str] = None

    # raw requirements / imports that are mapped
    raw: Optional[List[str]] = None

    # requirements mapping
    env: str = DEFAULT_REQUIREMENTS_ENV

    # output
    output_mode: str
    output_file: str

    # !!!NB!!! DO NOT USE DIRECTLY! INSTEAD, USE `get_output_extras_name`
    output_name: Optional[str] = None

    def get_output_extras_name(self) -> str:
        if self.output_name is not None:
            name = self.output_name
        elif self.start_scope is not None:
            name = self.start_scope
        elif self.scope is not None:
            name = self.scope
        else:
            raise ValueError(
                f"output_name cannot be determined, please set output_name, start_scope, or scope for: {self}"
            )
        return normalize_extras_name(name, strict=False)

    def get_manual_imports(self):
        if not self.raw:
            return []
        return [ManualImportInfo.from_target(r) for r in self.raw]

    @pydantic.field_validator("output_name", mode="before")
    @classmethod
    def _validate_output_name(cls, v):
        # TODO: should maybe allow non-strict mode?
        if v is not None:
            return normalize_extras_name(v, strict=True)
        return v

    @pydantic.field_validator("raw", mode="before")
    @classmethod
    def _validate_raw(cls, v):
        if v is not None:
            normalized = []
            for r in v:
                normalized.append(normalize_pkg_name(r, strict=False))
            return normalized
        return v

    @pydantic.model_validator(mode="after")
    @classmethod
    def _validate_model(cls, v):
        if v.start_scope is not None:
            if v.scope is None:
                raise ValueError(f"start_scope is set, but scope is not set for: {v}")
        return v

    def get_resolved_imports(
        self,
        loaded_scopes: "LoadedScopes",
    ):
        if not self.scope:
            return []
        # * normal scope
        if self.scope not in loaded_scopes:
            raise ValueError(
                f"scope {repr(self.scope)} does not exist, must be one of: {loaded_scopes.sorted_names}"
            )
        else:
            scope = loaded_scopes[self.scope]
        # * start scope
        start_scope = None
        if self.start_scope:
            if self.start_scope not in loaded_scopes:
                raise ValueError(
                    f"start_scope {repr(self.start_scope)} does not exist, must be one of: {loaded_scopes.sorted_names}"
                )
            else:
                start_scope = loaded_scopes[self.start_scope]
        # * resolve imports
        return scope.resolve_imports(
            start_scope=start_scope,
            visit_lazy=self.visit_lazy,
            re_add_lazy=self.re_add_lazy,
            exclude_unvisited=self.exclude_unvisited,
            exclude_in_search_space=self.exclude_in_search_space,
            exclude_builtins=self.exclude_builtins,
        )

    def resolve_generate_and_write_requirements(
        self,
        loaded_scopes: "LoadedScopes",
        requirements_mapper: RequirementsMapper,
    ) -> None:
        # 1. resolve imports
        resolved_imports = self.get_resolved_imports(loaded_scopes=loaded_scopes)
        manual_imports = self.get_manual_imports()
        # 2. generate requirements
        try:
            mapped_requirements = requirements_mapper.generate_output_requirements(
                imports=resolved_imports + manual_imports,
                requirements_env=self.env,
                strict=self.strict_requirements_map,
                resolver_name=self.get_output_extras_name(),
            )
        except NoConfiguredRequirementMappingError as e:
            msg = f"\n  | ".join(["", *str(e).split("\n")])
            msg = f"[requirement-mapping-error] output: {self.get_output_extras_name()}{msg}"
            raise NoConfiguredRequirementMappingError(msg, e.imports) from e
        # 3. write requirements
        self._write_requirements(
            mapped_requirements=mapped_requirements,
        )

    def _write_requirements(self, mapped_requirements: OutMappedRequirements) -> None:
        raise NotImplementedError(
            f"tried to write imports for {repr(self.get_output_extras_name())}, write_imports not implemented for {self.__class__.__name__}"
        )


class _OutputRequirements(_Output):
    output_mode: Literal[OutputModeEnum.requirements]

    def _write_requirements(self, mapped_requirements: OutMappedRequirements):
        string = mapped_requirements.as_requirements_txt(
            notice=True,
            sources=True,
            sources_compact=False,
            sources_roots=False,
            indent_size=4,
        )
        LOGGER.info(f"writing requirements to: {self.output_file}")
        txt_file_dump(
            file=self.output_file,
            contents=string,
        )


class _OutputPyprojectOptionalDeps(_Output):
    output_mode: Literal[OutputModeEnum.optional_dependencies]
    output_file: Optional[str] = None

    def _write_requirements(self, mapped_requirements: OutMappedRequirements):
        array = mapped_requirements.as_toml_array(
            notice=True,
            sources=True,
            sources_compact=False,
            sources_roots=False,
            indent_size=4,
        )
        out_name = self.get_output_extras_name()
        LOGGER.info(
            f"writing optional dependencies: {repr(out_name)} to: {self.output_file}"
        )
        toml_file_replace_array(
            file=self.output_file,
            keys=["project", "optional-dependencies", out_name],
            array=array,
        )


class _OutputPyprojectDeps(_Output):
    output_mode: Literal[OutputModeEnum.dependencies]
    output_file: Optional[str] = None

    def _write_requirements(self, mapped_requirements: OutMappedRequirements):
        array = mapped_requirements.as_toml_array(
            notice=True,
            sources=True,
            sources_compact=False,
            sources_roots=False,
            indent_size=4,
        )
        LOGGER.info(f"writing dependencies to: {self.output_file}")
        toml_file_replace_array(
            file=self.output_file,
            keys=["project", "dependencies"],
            array=array,
        )


CfgResolver = Annotated[
    Union[
        _OutputRequirements,
        _OutputPyprojectOptionalDeps,
        _OutputPyprojectDeps,
    ],
    pydantic.Field(discriminator="output_mode", union_mode="left_to_right"),
]

# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #
# CONFIG - PACKAGES                                                         #
# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #


class InvalidRequirementsName(ValueError):
    pass


class InvalidExtrasName(ValueError):
    pass


def normalize_pkg_name(string: str, strict: bool = True):
    """
    Ensure uses "_" instead of "-", and no whitespace. Also convert to lowercase.
    """
    norm = string.replace("-", "_").lower()
    if any(c in norm for c in [" ", "\t", "\n"]):
        raise InvalidRequirementsName(f"Requirements name is invalid: {repr(string)}")
    if norm != string:
        if strict:
            raise InvalidRequirementsName(
                f"Requirements name is invalid: {repr(string)}, should be: {repr(norm)}"
            )
        else:
            warnings.warn(
                f"normalized requirements name from {repr(string)} to {repr(norm)}"
            )
    return norm


def normalize_extras_name(string: str, strict: bool = True):
    """
    Ensure uses "-" instead of "_", and no whitespace. Keeps case.
    """
    norm = string.replace("_", "-")
    if any(c in norm for c in [" ", "\t", "\n"]):
        raise InvalidExtrasName(f"Extras name is invalid: {repr(string)}")
    if norm != string:
        if strict:
            raise InvalidExtrasName(
                f"Extras name is invalid: {repr(string)}, should be: {repr(norm)}"
            )
        else:
            warnings.warn(f"normalized extras name from {repr(string)} to {repr(norm)}")
    return norm


class CfgVersion(pydantic.BaseModel, extra="forbid", arbitrary_types_allowed=True):
    # the pip install requirement
    requirement: str
    # the imports to replace
    import_: Optional[List[str]] = pydantic.Field(default=None, alias="import")
    scope: Optional[str] = None
    # only apply this import to this environment
    env: str = DEFAULT_REQUIREMENTS_ENV

    @property
    def parsed_requirement(self) -> Requirement:
        return Requirement(self.requirement)

    @property
    def package(self) -> str:
        return normalize_pkg_name(self.parsed_requirement.name, strict=False)

    @classmethod
    def from_string(cls, requirement: str):
        return cls(requirement=requirement)

    def get_import_matcher(self, loaded_scopes: "LoadedScopes") -> ImportMatcherBase:
        if self.scope is not None:
            if self.import_ is not None:
                raise ValueError(f"cannot specify both scope and import for: {self}")
            else:
                return ImportMatcherScope(scope=loaded_scopes[self.scope])
        else:
            if self.import_ is None:
                raise ValueError(f"must specify either scope or import for: {self}")
            else:
                return ImportMatcherGlobs(import_globs=self.import_)

    @pydantic.model_validator(mode="after")
    @classmethod
    def _validate_model_before(cls, v: "CfgVersion"):
        if not str.isidentifier(v.env.replace("-", "_")):
            raise ValueError(
                f"env must be a valid identifier (with hyphens replaced with underscores), got: {v.env}"
            )
        if v.import_ is None and v.scope is None:
            v.import_ = [f"{v.package}.*"]  # wildcard
        elif v.import_ is not None and v.scope is not None:
            raise ValueError(f"cannot specify both scope and import for: {v}")
        return v

    @pydantic.field_validator("import_", mode="before")
    @classmethod
    def _validate_import(cls, v):
        if v is not None:
            if isinstance(v, str):
                v = v.split(",")
        return v


# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #
# CONFIG - SCOPE                                                            #
# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #


class _ScopeRules(pydantic.BaseModel, extra="forbid"):

    # Specify how to handle modules that are unreachable, e.g. if there is no `__init__.py`
    # file in all the parents leading up to importing this module. If this is the case
    # then the module/package does not correctly follow python/PEP convention and is
    # technically invalid. By default, for `error`, we raise an exception and do not allow
    # the scope to be created, but this can be relaxed to `skip` or `keep` these files.
    unreachable_mode: Optional[UnreachableModeEnum] = None

    @classmethod
    def make_default_base_rules(cls):
        return _ScopeRules(
            unreachable_mode=UnreachableModeEnum.error,
        )

    def set_defaults(self, defaults: "_ScopeRules"):
        assert defaults.unreachable_mode is not None
        if self.unreachable_mode is None:
            self.unreachable_mode = defaults.unreachable_mode


class CfgScope(_ScopeRules, extra="forbid"):
    # name
    # - must be unique across all scopes & sub-scopes
    name: str

    # parents
    parents: List[str] = pydantic.Field(default_factory=list)

    # search paths
    search_paths: List[str] = pydantic.Field(default_factory=list)
    pkg_paths: List[str] = pydantic.Field(default_factory=list)
    unreachable_mode: Optional[UnreachableModeEnum] = None

    # extra packages
    # packages: List[str] = pydantic.Field(default_factory=list)

    # filtering: limit > exclude > [include!!] (in that order)
    # - order is important because it allows us to remove a band of modules
    #   e.g. limit=foo.bar, exclude=foo.bar.baz, include=foo.bar.baz.qux
    #   if order of include and exclude were swapped, then the exclude would
    #   remove the module after the include added it back in
    limit: Optional[List[str]] = None
    exclude: Optional[List[str]] = None
    # include: Optional[str] = None  # NOT IMPLEMENTED BECAUSE IT IS REDUNDANT, AND `PARENTS` CAN BE USED INSTEAD

    # sub-scopes
    # - name to import path map
    # - names must be unique across all scopes & sub-scopes
    # - imports must belong to the scope
    subscopes: Dict[str, str] = pydantic.Field(default_factory=dict)

    @pydantic.field_validator("search_paths", mode="before")
    @classmethod
    def _validate_search_paths(cls, v):
        return [v] if isinstance(v, str) else v

    @pydantic.field_validator("pkg_paths", mode="before")
    @classmethod
    def _validate_pkg_paths(cls, v):
        return [v] if isinstance(v, str) else v

    @pydantic.field_validator("limit", mode="before")
    @classmethod
    def _validate_limit(cls, v):
        return [v] if isinstance(v, str) else v

    @pydantic.field_validator("exclude", mode="before")
    @classmethod
    def _validate_exclude(cls, v):
        return [v] if isinstance(v, str) else v

    def make_module_scope(self, loaded_scopes: "LoadedScopes" = None):
        m = ModulesScope()

        # 1. load parents
        if self.parents:
            if loaded_scopes is None:
                raise ValueError("loaded_scopes must be provided if parents are used!")
            for parent in self.parents:
                if parent not in loaded_scopes:
                    raise ValueError(
                        f"parent scope {repr(parent)} has not yet been created, are you sure the order of definitions is correct?"
                    )
                m.add_modules_from_scope(loaded_scopes[parent])

        # 2. load new search paths and packages
        for path in self.search_paths:
            m.add_modules_from_search_path(
                Path(path),
                tag=self.name,
                unreachable_mode=self.unreachable_mode,
            )
        for path in self.pkg_paths:
            m.add_modules_from_package_path(
                Path(path),
                tag=self.name,
                unreachable_mode=self.unreachable_mode,
            )

        # 3. add extra packages
        # if self.packages:
        # m.add_modules_from_raw_imports(
        #     imports=self.packages,
        #     tag=self.name,
        # )
        # raise NotImplementedError("extra packages not yet implemented!")

        # 4. filter everything
        # - a. limit, b. exclude, [c. include (replaced with parents)]
        if self.limit:
            m = m.get_restricted_scope(
                imports=self.limit,
                mode=RestrictMode.CHILDREN,
                op=RestrictOp.LIMIT,
            )
        if self.exclude:
            m = m.get_restricted_scope(
                imports=self.exclude,
                mode=RestrictMode.CHILDREN,
                op=RestrictOp.EXCLUDE,
            )

        # done!
        return m


# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #
# Loaded Scopes                                                             #
# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #


class UndefinedScopeError(ValueError):
    pass


class LoadedScopes:

    def __init__(self):
        self._scopes = {}

    def __contains__(self, item):
        return item in self._scopes

    def __getitem__(self, item: str) -> ModulesScope:
        if item not in self._scopes:
            raise UndefinedScopeError(
                f"scope {repr(item)} is not defined, must be one of: {self.sorted_names}"
            )
        return self._scopes[item]

    def __setitem__(self, key, value):
        assert isinstance(value, ModulesScope)
        if key in self:
            raise ValueError(f"scope {repr(key)} is already defined!")
        self._scopes[key] = value

    @property
    def sorted_names(self) -> List[str]:
        return sorted(self._scopes.keys())


# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #
# CONFIG - ROOT                                                             #
# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #


class PydependenceCfg(pydantic.BaseModel, extra="forbid"):
    # default root is relative to the parent of the pyproject.toml file
    # and is the folder containing the repo of the pyproject.toml file
    default_root: str = "."

    # default write modes
    default_resolve_rules: _ResolveRules = pydantic.Field(
        default_factory=_ResolveRules.make_default_base_rules
    )
    default_scope_rules: _ScopeRules = pydantic.Field(
        default_factory=_ScopeRules.make_default_base_rules
    )

    # package versions
    versions: List[CfgVersion] = pydantic.Field(default_factory=list)

    # resolve
    scopes: List[CfgScope] = pydantic.Field(default_factory=dict)

    # outputs
    resolvers: List[CfgResolver] = pydantic.Field(default_factory=list)

    @pydantic.field_validator("versions", mode="before")
    @classmethod
    def _validate_versions(cls, v, values):
        versions = []
        reqs_envs = set()  # pairs of tags and req names must be unique for now
        for x in v:
            if isinstance(x, str):
                x = CfgVersion.from_string(x)
            else:
                x = CfgVersion.model_validate(x)
            req_env = (x.package, x.env)
            if req_env in reqs_envs:
                raise ValueError(
                    f"requirement {repr(x.package)} and env {repr(x.env)} combination is defined multiple times! ({repr(x.requirement)})"
                )
            reqs_envs.add(req_env)
            versions.append(x)
        return versions

    @pydantic.model_validator(mode="after")
    @classmethod
    def _validate_model(cls, cfg: "PydependenceCfg"):
        # 1. check that scope names are all unique
        scope_names = set()
        for scope in cfg.scopes:
            if scope.name in scope_names:
                raise ValueError(f"scope name {repr(scope.name)} is not unique!")
            scope_names.add(scope.name)

        # 2. check that all sub-scope names are unique
        for scope in cfg.scopes:
            for subscope_name in scope.subscopes:
                if subscope_name in scope_names:
                    raise ValueError(
                        f"sub-scope name {repr(subscope_name)} is not unique!"
                    )
                scope_names.add(subscope_name)

        # 3. check that all packages
        # TODO

        # 4. check that the default root is a relative path
        if Path(cfg.default_root).is_absolute():
            raise ValueError(
                f"default_root must be a relative path, got: {repr(cfg.default_root)}"
            )
        return cfg

    def apply_defaults(self, *, config_path: "Union[str, Path]"):
        """
        config_path is the path to the pyproject.toml file or the toml file that was
        used to load the configuration. This is used to determine the default root path,
        which is the folder containing the config. The default root is then used to
        resolve all relative paths in the configuration.
        """
        config_path = Path(config_path)

        # helper
        self.default_root = apply_root_to_path_str(
            config_path.parent, self.default_root
        )

        def _resolve_path(x: "Union[str, Path]") -> str:
            return apply_root_to_path_str(self.default_root, x)

        # apply to all paths
        for scope in self.scopes:
            scope.search_paths = [_resolve_path(x) for x in scope.search_paths]
            scope.pkg_paths = [_resolve_path(x) for x in scope.pkg_paths]
        for output in self.resolvers:
            if output.output_file is not None:
                output.output_file = _resolve_path(output.output_file)
            if output.output_file is None:
                if isinstance(
                    output, (_OutputPyprojectDeps, _OutputPyprojectOptionalDeps)
                ):
                    output.output_file = _resolve_path(config_path)
            # check kinds
            if isinstance(output, (_OutputPyprojectDeps, _OutputPyprojectOptionalDeps)):
                if Path(output.output_file).name != "pyproject.toml":
                    raise ValueError(
                        f"output_file must be the pyproject.toml file for: {output}"
                    )
            elif isinstance(output, _OutputRequirements):
                if Path(output.output_file).suffix != ".txt":
                    raise ValueError(
                        f"output_file must be requirements*.txt for: {output}"
                    )

        # also apply all default write modes
        self.default_scope_rules.set_defaults(_ScopeRules.make_default_base_rules())
        self.default_resolve_rules.set_defaults(_ResolveRules.make_default_base_rules())
        for scope in self.scopes:
            scope.set_defaults(self.default_scope_rules)
        for output in self.resolvers:
            output.set_defaults(self.default_resolve_rules)

    def load_scopes(self) -> "LoadedScopes":
        # resolve all scopes
        loaded_scopes = LoadedScopes()
        for scope_cfg in self.scopes:
            scope = scope_cfg.make_module_scope(loaded_scopes=loaded_scopes)
            loaded_scopes[scope_cfg.name] = scope
            # now create sub-scopes
            for subcol_name, subcol_import_root in scope_cfg.subscopes.items():
                subscope = scope.get_restricted_scope(
                    imports=[subcol_import_root], mode=RestrictMode.CHILDREN
                )
                loaded_scopes[subcol_name] = subscope
        # done!
        return loaded_scopes

    def make_requirements_mapper(
        self,
        loaded_scopes: "LoadedScopes",
    ):
        env_matchers = defaultdict(list)
        for v in self.versions:
            import_matcher = v.get_import_matcher(loaded_scopes=loaded_scopes)
            pair = ReqMatcher(requirement=v.requirement, matcher=import_matcher)
            env_matchers[v.env].append(pair)
        env_matchers = dict(env_matchers)

        return RequirementsMapper(
            env_matchers=env_matchers,
        )

    def write_all_outputs(self, loaded_scopes: "LoadedScopes"):
        # check that scope output names are unique
        # - output names only need to be unique if they are optional-dependencies!
        # - warn if generally not unique, error if optional-deps not unique
        names_all = set()
        names_optional_deps = set()
        for output in self.resolvers:
            name = output.get_output_extras_name()
            if name in names_all:
                warnings.warn(
                    f"output name {repr(name)} is not unique across all resolvers!"
                )
            names_all.add(name)
            if output.output_mode == OutputModeEnum.optional_dependencies:
                if name in names_optional_deps:
                    raise ValueError(
                        f"output name {repr(name)} is not unique across resolvers for optional dependencies!"
                    )
                names_optional_deps.add(name)

        # check that the scopes exists
        for output in self.resolvers:
            if output.scope is None:
                assert output.start_scope is None
                continue
            if output.scope not in loaded_scopes:
                raise ValueError(
                    f"output scope {repr(output.scope)} does not exist! Are you sure it has been defined? Available scopes: {loaded_scopes.sorted_names}"
                )
            if output.start_scope and output.start_scope not in loaded_scopes:
                raise ValueError(
                    f"output start_scope {repr(output.start_scope)} does not exist! Are you sure it has been defined? Available scopes: {loaded_scopes.sorted_names}"
                )

        # make the mapper
        requirements_mapper = self.make_requirements_mapper(loaded_scopes=loaded_scopes)

        # resolve the scopes!
        for output in self.resolvers:
            output.resolve_generate_and_write_requirements(
                loaded_scopes=loaded_scopes,
                requirements_mapper=requirements_mapper,
            )

    # ... LOADING ...

    @classmethod
    def from_pyproject(cls, path: Path) -> "PydependenceCfg":
        # 1. load pyproject.toml
        toml = load_toml_document(path)
        # 2. validate the model
        pyproject = _PyprojectToml.model_validate(toml.unwrap())
        pydependence = pyproject.tool.pydependence
        # 3. override paths in cfg using the default root
        pydependence.apply_defaults(config_path=path)
        return pydependence

    @classmethod
    def from_toml_config(cls, path: Path) -> "PydependenceCfg":
        # 1. load pyproject.toml
        toml = load_toml_document(path)
        # 2. validate the model
        tool = _PyprojectTomlTools.model_validate(toml.unwrap())
        pydependence = tool.pydependence
        # 3. override paths in cfg using the default root
        pydependence.apply_defaults(config_path=path)
        return pydependence

    @classmethod
    def from_file_automatic(cls, path: Path) -> "PydependenceCfg":
        if path.name == "pyproject.toml":
            return cls.from_pyproject(path)
        elif path.name == ".pydependence.toml":
            return cls.from_toml_config(path)
        elif path.suffix in (".toml", ".cfg"):
            LOGGER.warning(
                f"using legacy extension mode: {repr(path.suffix)}, not explicit file name as one of: 'pyproject.toml' OR '.pydependence.toml'"
            )
            return cls.from_toml_config(path)
        else:
            raise ValueError(f"unsupported file extension: {path.suffix} for: {path}")


# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #
# CONFIG - PYPROJECT                                                        #
# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #


class _PyprojectTomlTools(pydantic.BaseModel, extra="ignore"):
    pydependence: PydependenceCfg


class _PyprojectToml(pydantic.BaseModel, extra="ignore"):
    tool: _PyprojectTomlTools = pydantic.Field(default_factory=_PyprojectTomlTools)


# ========================================================================= #
# COLLECT MODULES                                                           #
# ========================================================================= #


def pydeps(
    *,
    config_path: Union[str, Path],
):
    # 1. get absolute
    config_path = Path(config_path).resolve().absolute()
    LOGGER.info(f"loading pydependence config from: {config_path}")
    # 2. load pyproject.toml
    pydependence = PydependenceCfg.from_file_automatic(config_path)
    # 3. generate search spaces, recursively resolving!
    loaded_scopes = pydependence.load_scopes()
    # 4. generate outputs
    pydependence.write_all_outputs(loaded_scopes)


# ========================================================================= #
# END                                                                       #
# ========================================================================= #



# ============================================================================== #
# MIT License                                                                    #
#                                                                                #
# Copyright (c) 2024 Nathan Juraj Michlo                                         #
#                                                                                #
# Permission is hereby granted, free of charge, to any person obtaining a copy   #
# of this software and associated documentation files (the "Software"), to deal  #
# in the Software without restriction, including without limitation the rights   #
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell      #
# copies of the Software, and to permit persons to whom the Software is          #
# furnished to do so, subject to the following conditions:                       #
#                                                                                #
# The above copyright notice and this permission notice shall be included in all #
# copies or substantial portions of the Software.                                #
#                                                                                #
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR     #
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,       #
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE    #
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER         #
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,  #
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE  #
# SOFTWARE.                                                                      #
# ============================================================================== #


import dataclasses
from typing import Dict, List, Tuple

from pydependence._core.module_data import ModuleMetadata
from pydependence._core.module_imports_ast import (
    LocImportInfo,
    load_imports_from_module_info,
)

# ========================================================================= #
# MODULE IMPORTS                                                            #
# ========================================================================= #


@dataclasses.dataclass
class ModuleImports:
    module_info: ModuleMetadata
    module_imports: "Dict[str, List[LocImportInfo]]"

    @classmethod
    def from_module_info_and_parsed_file(cls, module_info: ModuleMetadata):
        module_imports = load_imports_from_module_info(module_info=module_info)
        return ModuleImports(
            module_info=module_info,
            module_imports=dict(module_imports),
        )


# ========================================================================= #
# MODULE IMPORTS LOADER                                                     #
# ========================================================================= #


class _ModuleImportsLoader:

    def __init__(self):
        # TODO: tag could severally hurt performance? maybe should change data structure slightly?
        #       problem is tag is nested and applied to imports too. HOWEVER, Usually tag is
        #       automatically generated from the package name, so might not matter too much in practice.
        self._modules_imports: "Dict[Tuple[str, str], ModuleImports]" = {}

    def load_module_imports(self, module_info: ModuleMetadata) -> ModuleImports:
        k = (module_info.name, module_info.tag)
        v = self._modules_imports.get(k, None)
        if v is None:
            v = ModuleImports.from_module_info_and_parsed_file(module_info)
            self._modules_imports[k] = v
        else:
            if v.module_info != module_info:
                raise RuntimeError(
                    f"ModuleMetadata mismatch: {v.module_info} != {module_info}"
                )
        return v


# GLOBAL INSTANCE
# TODO: can replace with disk cache
DEFAULT_MODULE_IMPORTS_LOADER = _ModuleImportsLoader()


# ========================================================================= #
# END                                                                       #
# ========================================================================= #


__all__ = (
    "ModuleImports",
    "DEFAULT_MODULE_IMPORTS_LOADER",
)



# ============================================================================== #
# MIT License                                                                    #
#                                                                                #
# Copyright (c) 2024 Nathan Juraj Michlo                                         #
#                                                                                #
# Permission is hereby granted, free of charge, to any person obtaining a copy   #
# of this software and associated documentation files (the "Software"), to deal  #
# in the Software without restriction, including without limitation the rights   #
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell      #
# copies of the Software, and to permit persons to whom the Software is          #
# furnished to do so, subject to the following conditions:                       #
#                                                                                #
# The above copyright notice and this permission notice shall be included in all #
# copies or substantial portions of the Software.                                #
#                                                                                #
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR     #
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,       #
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE    #
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER         #
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,  #
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE  #
# SOFTWARE.                                                                      #
# ============================================================================== #
import abc
import dataclasses
import functools
import warnings
from typing import TYPE_CHECKING, Dict, List, NamedTuple, Optional, Set, Union

from pydependence._core.builtin import BUILTIN_MODULE_NAMES
from pydependence._core.module_imports_ast import (
    BasicImportInfo,
    LocImportInfo,
    ManualImportInfo,
)
from pydependence._core.requirements_out import (
    OutMappedRequirement,
    OutMappedRequirements,
    OutMappedRequirementSource,
)

if TYPE_CHECKING:
    from pydependence._core.modules_scope import ModulesScope


# The version matching env
DEFAULT_REQUIREMENTS_ENV = "default"


# ========================================================================= #
# IMPORT MATCHER                                                            #
# ========================================================================= #


class ImportMatcherBase(abc.ABC):

    @abc.abstractmethod
    def match(self, import_: str) -> bool:
        raise NotImplementedError

    @abc.abstractmethod
    def cfg_str(self) -> str:
        raise NotImplementedError


class ImportMatcherScope(ImportMatcherBase):

    def __init__(self, scope: "ModulesScope"):
        self.scope = scope

    def match(self, import_: str) -> bool:
        return self.scope.has_module(import_)

    def cfg_str(self) -> str:
        return f"scope={repr(self.scope)}"


class ImportMatcherGlob(ImportMatcherBase):

    def __init__(self, import_glob: str):
        self._orig = import_glob
        (*parts, last) = import_glob.split(".")
        # check all parts are identifiers, OR, at least one identifier with the last part being a glob
        if parts:
            if not all(str.isidentifier(x) for x in parts):
                raise ValueError(
                    f"parts of import glob {repr(import_glob)} are not valid identifiers"
                )
            if not (str.isidentifier(last) or last == "*"):
                raise ValueError(
                    f"last part of import glob {repr(import_glob)} is not a valid identifier or '*'"
                )
        else:
            if not str.isidentifier(last):
                raise ValueError(
                    f"last part of import glob {repr(import_glob)} is not a valid identifier"
                )
        # create glob
        if last == "*":
            self._parts = parts
            self._wildcard = True
        else:
            self._parts = (*parts, last)
            self._wildcard = False
        self._base = ".".join(self._parts)

    def match(self, import_: str) -> bool:
        if not self._wildcard:
            return import_ == self._base
        else:
            parts = import_.split(".")
            return self._parts == parts[: len(self._parts)]

    def cfg_str(self) -> str:
        return f"import={repr(self._orig)}"


class ImportMatcherGlobs(ImportMatcherBase):

    def __init__(self, import_globs: "Union[str, List[str]]"):
        if isinstance(import_globs, str):
            import_globs = import_globs.split(",")
        self._orig = ",".join(import_globs)
        # create
        self._matchers = []
        # dedupe
        _added = set()
        for x in import_globs:
            if x not in _added:
                self._matchers.append(ImportMatcherGlob(x))
                _added.add(x)

    def match(self, import_: str) -> bool:
        # linear search... inefficient...
        for matcher in self._matchers:
            if matcher.match(import_):
                return True
        return False

    def cfg_str(self) -> str:
        return f"import={repr(self._orig)}"


# ========================================================================= #
# REQUIREMENTS MAPPER (INFO)                                                #
# ========================================================================= #


class MappedRequirementInfo(NamedTuple):
    requirement: str
    is_mapped: bool
    original_name: str


# ========================================================================= #
# REQUIREMENTS MAPPER (INTERMEDIATE)                                        #
# # similar data structures to output, but with more information
# ========================================================================= #


@dataclasses.dataclass
class MappedRequirementSource:
    source_module: str
    source_module_imports: List[BasicImportInfo]

    def to_output_requirement_source(self):
        return OutMappedRequirementSource(
            source_module=self.source_module,
            is_lazy=all(imp.is_lazy for imp in self.source_module_imports),
            is_manual=any(
                isinstance(imp, ManualImportInfo) for imp in self.source_module_imports
            ),
        )


@dataclasses.dataclass
class MappedRequirement:
    requirement: str  # mapped name
    sources: Dict[str, MappedRequirementSource]  # k == v.source_module

    def get_sorted_sources(self) -> List[MappedRequirementSource]:
        return sorted(self.sources.values(), key=lambda x: x.source_module)

    def to_output_requirement(self):
        return OutMappedRequirement(
            requirement=self.requirement,
            sources=[
                source.to_output_requirement_source()
                for source in self.get_sorted_sources()
            ],
        )


@dataclasses.dataclass
class MappedRequirements:
    requirements: Dict[str, MappedRequirement]  # k == v.requirement
    resolver_name: Optional[str] = None

    def get_sorted_requirements(self) -> List[MappedRequirement]:
        return sorted(
            self.requirements.values(),
            key=lambda x: x.requirement,
        )

    def to_output_requirements(self):
        return OutMappedRequirements(
            requirements=[
                requirement_info.to_output_requirement()
                for requirement_info in self.get_sorted_requirements()
            ],
            resolver_name=self.resolver_name,
        )


# ========================================================================= #
# REQUIREMENTS MAPPER                                                       #
# ========================================================================= #


class NoConfiguredRequirementMappingError(ValueError):

    def __init__(self, msg: str, imports: Set[str]):
        self.msg = msg
        self.imports = imports
        super().__init__(msg)


@dataclasses.dataclass(frozen=True)
class ReqMatcher:
    requirement: str
    matcher: ImportMatcherBase

    def cfg_str(self) -> str:
        return f"{{requirement={repr(self.requirement)}, {self.matcher.cfg_str()}}}"


class RequirementsMapper:

    def __init__(
        self,
        *,
        env_matchers: "Optional[Union[Dict[str, List[ReqMatcher]], List[ReqMatcher]]]",
    ):
        # env -> [(requirement, import matcher), ...]
        # * we use a list to maintain order, and then linear search. This is because
        #   we could have multiple imports that match to the same requirement.
        #   we could potentially be stricter about this in future...
        self._env_matchers = self._validate_env_matchers(env_matchers)

    @classmethod
    def _validate_env_matchers(cls, env_matchers) -> "Dict[str, List[ReqMatcher]]":
        # normalize
        if env_matchers is None:
            env_matchers = {}
        elif not isinstance(env_matchers, dict):
            env_matchers = {DEFAULT_REQUIREMENTS_ENV: list(env_matchers)}

        # shift
        if None in env_matchers:
            if DEFAULT_REQUIREMENTS_ENV in env_matchers:
                raise ValueError(
                    f"env_matchers cannot have both {repr(None)} and {repr(DEFAULT_REQUIREMENTS_ENV)} as keys."
                )
            env_matchers[DEFAULT_REQUIREMENTS_ENV] = env_matchers.pop(None)

        # check
        if not isinstance(env_matchers, dict):
            raise ValueError(
                f"env_matchers must be a dictionary, got: {type(env_matchers)}"
            )
        for env, matchers in env_matchers.items():
            if not isinstance(matchers, list):
                raise ValueError(
                    f"env_matchers must be a dictionary of lists, got: {type(matchers)}"
                )
            for matcher in matchers:
                if not isinstance(matcher, ReqMatcher):
                    raise ValueError(
                        f"env_matchers must be a dictionary of lists of ReqMatcherPair, got: {type(matcher)}, {matcher}"
                    )
                if not isinstance(matcher.requirement, str):
                    raise ValueError(
                        f"requirement must be a string, got: {type(matcher.requirement)}"
                    )
                if not isinstance(matcher.matcher, ImportMatcherBase):
                    raise ValueError(
                        f"matcher must be an ImportMatcherBase, got: {type(matcher.matcher)}, {matcher.matcher}"
                    )
        return env_matchers

    def map_import_to_requirement(
        self,
        import_: str,
        *,
        requirements_env: "Optional[str]" = None,
        strict: bool = False,
    ) -> str:
        req_info = self.map_import_to_requirement_info(
            import_,
            requirements_env=requirements_env,
            strict=strict,
        )
        return req_info.requirement

    @functools.lru_cache(maxsize=256)
    def map_import_to_requirement_info(
        self,
        import_: str,
        *,
        requirements_env: "Optional[str]" = None,
        strict: bool = False,
    ) -> "MappedRequirementInfo":
        """
        :raises NoConfiguredRequirementMappingError: if no requirement is found for an import and if strict mode is enabled.
        """
        if requirements_env is None:
            requirements_env = DEFAULT_REQUIREMENTS_ENV
        # 1. take the specific env
        if requirements_env != DEFAULT_REQUIREMENTS_ENV:
            if requirements_env not in self._env_matchers:
                raise ValueError(
                    f"env: {repr(requirements_env)} has not been defined for a requirement."
                )
            for rm in self._env_matchers[requirements_env]:
                if rm.matcher.match(import_):
                    return MappedRequirementInfo(
                        rm.requirement,
                        is_mapped=True,
                        original_name=import_,
                    )
        # 2. take the default env
        for rm in self._env_matchers.get(DEFAULT_REQUIREMENTS_ENV, []):
            if rm.matcher.match(import_):
                return MappedRequirementInfo(
                    rm.requirement,
                    is_mapped=True,
                    original_name=import_,
                )
        # 3. return the root
        if strict:
            raise NoConfiguredRequirementMappingError(
                msg=f"could not find import to requirement mappings: {repr(import_)},\ndefine a scope or glob matcher for this import, or set disable strict mode!",
                imports={import_},
            )
        else:
            root = import_.split(".")[0]
            warnings.warn(
                f"could not find a matching requirement for import: {repr(import_)}, returning the import root: {repr(root)} as the requirement"
            )
            return MappedRequirementInfo(
                root,
                is_mapped=False,
                original_name=import_,  # TODO: or should this be root?
            )

    def _get_joined_matchers(
        self,
        requirements_env: "Optional[str]" = None,
    ) -> "List[ReqMatcher]":
        if requirements_env is None:
            requirements_env = DEFAULT_REQUIREMENTS_ENV
        matchers = self._env_matchers.get(requirements_env, [])
        if requirements_env != DEFAULT_REQUIREMENTS_ENV:
            matchers = self._env_matchers.get(DEFAULT_REQUIREMENTS_ENV, []) + matchers
        return matchers

    def _get_matcher_cfg_sting(self, requirements_env: "Optional[str]" = None) -> "str":
        return ", ".join(
            [
                rm.cfg_str()
                for rm in self._get_joined_matchers(requirements_env=requirements_env)
            ]
        )

    def generate_mapped_requirements(
        self,
        imports: "List[BasicImportInfo]",
        *,
        requirements_env: "Optional[str]" = None,
        strict: bool = False,
        raw: List[str] = None,
        resolver_name: Optional[str] = None,
    ) -> "MappedRequirements":
        """
        Map imports to requirements, returning the imports grouped by the requirement.

        :raises NoConfiguredRequirementMappingError: if no requirement is found for an import, but only if strict mode is enabled, and after all imports have been processed so that pretty error messages can be generated.
        """
        if raw:
            raise NotImplementedError("raw imports are not yet supported, TODO!!!")

        # group imports by requirement
        r = MappedRequirements(
            requirements={},
            resolver_name=resolver_name,
        )
        errors = []
        for imp in imports:
            # 1. map requirements
            if imp.target in BUILTIN_MODULE_NAMES:  # TODO: needed?
                req_info = MappedRequirementInfo(
                    imp.target,
                    is_mapped=False,
                    original_name=imp.target,
                )
            elif imp.root_target in BUILTIN_MODULE_NAMES:
                req_info = MappedRequirementInfo(
                    imp.root_target,
                    is_mapped=False,
                    original_name=imp.target,  # TODO: or should this be root?
                )
            else:
                try:
                    req_info = self.map_import_to_requirement_info(
                        imp.target,
                        requirements_env=requirements_env,
                        strict=strict,
                    )
                except NoConfiguredRequirementMappingError as e:
                    errors.append(e)
                    continue

            # 2. get or create requirement sources
            req_group = r.requirements.get(req_info.requirement, None)
            if req_group is None:
                req_group = MappedRequirement(
                    requirement=req_info.requirement,
                    sources={},
                )
                r.requirements[req_info.requirement] = req_group

            # - get or create requirement source import
            req_group_source = req_group.sources.get(imp.source_name, None)
            if req_group_source is None:
                req_group_source = MappedRequirementSource(
                    source_module=imp.source_name,
                    source_module_imports=[],
                )
                req_group.sources[imp.source_name] = req_group_source

            # - append import to source & update
            req_group_source.source_module_imports.append(imp)

        if errors:
            err_imports = {imp for e in errors for imp in e.imports}
            err_roots = {imp.split(".")[0] for imp in err_imports}
            raise NoConfiguredRequirementMappingError(
                msg=(
                    f"could not find import to requirement mappings for roots:"
                    f"\n  * {', '.join(map(repr, map(str, sorted(set(err_roots)))))},"
                    f"\nor full imports:"
                    f"\n  * {', '.join(map(repr, map(str, sorted(set(err_imports)))))},"
                    f"\navailable matchers: {self._get_matcher_cfg_sting(requirements_env=requirements_env) or '<NONE>'},"
                    f"\notherwise if running from a config file, set strict_requirements_map=False to disable strict mode and use the root module name instead."
                ),
                imports=err_imports,
            )

        # done!
        return r

    def generate_output_requirements(
        self,
        imports: "List[BasicImportInfo]",
        *,
        requirements_env: "Optional[str]" = None,
        strict: bool = False,
        resolver_name: Optional[str] = None,
    ) -> "OutMappedRequirements":
        """
        :raises NoConfiguredRequirementMappingError: if no requirement is found for any import, but only if strict mode is enabled.
        """
        # 1. map imports to requirements
        r = self.generate_mapped_requirements(
            imports,
            requirements_env=requirements_env,
            strict=strict,
            resolver_name=resolver_name,
        )
        # 2. generate output requirements lists
        return r.to_output_requirements()


# ========================================================================= #
# END                                                                       #
# ========================================================================= #



# ============================================================================== #
# MIT License                                                                    #
#                                                                                #
# Copyright (c) 2024 Nathan Juraj Michlo                                         #
#                                                                                #
# Permission is hereby granted, free of charge, to any person obtaining a copy   #
# of this software and associated documentation files (the "Software"), to deal  #
# in the Software without restriction, including without limitation the rights   #
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell      #
# copies of the Software, and to permit persons to whom the Software is          #
# furnished to do so, subject to the following conditions:                       #
#                                                                                #
# The above copyright notice and this permission notice shall be included in all #
# copies or substantial portions of the Software.                                #
#                                                                                #
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR     #
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,       #
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE    #
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER         #
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,  #
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE  #
# SOFTWARE.                                                                      #
# ============================================================================== #


import dataclasses
from collections import defaultdict
from typing import List, NamedTuple, Optional, Tuple

# ========================================================================= #
# REQUIREMENTS MAPPER                                                       #
# ========================================================================= #


@dataclasses.dataclass
class OutMappedRequirementSource:
    source_module: str
    is_lazy: bool
    is_manual: bool

    @property
    def source_module_root(self):
        return self.source_module.split(".")[0]


class SrcInfo(NamedTuple):
    name: str
    comment: str

    @property
    def anno_str(self) -> str:
        if self.comment:
            return f"← {self.comment} {self.name}"
        else:
            return f"← {self.name}"


@dataclasses.dataclass
class OutMappedRequirement:
    requirement: str
    sources: List[OutMappedRequirementSource]

    @property
    def all_lazy(self) -> bool:
        return all(src.is_lazy for src in self.sources)

    @property
    def any_manual(self) -> bool:
        return any(src.is_manual for src in self.sources)

    def get_source_info(
        self,
        enabled: bool = True,
        roots: bool = False,
        annotate: bool = True,
    ) -> "List[SrcInfo]":
        if enabled:
            if roots:
                r = defaultdict(lambda: True)
                for src in self.sources:
                    r[src.source_module_root] &= src.is_lazy
                return [
                    SrcInfo(name=k, comment="[L]" if annotate and r[k] else "")
                    for k in sorted(r.keys())
                ]
            else:
                return [
                    SrcInfo(
                        name=src.source_module,
                        comment=f"[L]" if annotate and src.is_lazy else "",
                    )
                    for src in self.sources
                ]
        else:
            return []

    def get_sources_string(
        self,
        enabled: bool = True,
        roots: bool = False,
    ) -> str:
        return ", ".join(
            i.name
            for i in self.get_source_info(enabled=enabled, roots=roots, annotate=False)
        )

    def get_annotations_string(
        self,
        enabled: bool = True,
        comment: bool = False,
    ) -> str:
        items = []
        if enabled:
            if self.all_lazy:
                items.append("L")
            if self.any_manual:
                items.append("M")
        items = "".join(items)
        if items:
            return f" # [{items}]" if comment else f"[{items}]"
        return ""


@dataclasses.dataclass
class OutMappedRequirements:
    requirements: List[OutMappedRequirement]
    resolver_name: Optional[str] = None

    _AUTOGEN_NOTICE = "[AUTOGEN] by pydependence **DO NOT EDIT** [AUTOGEN]"
    _AUTOGEN_NOTICE_NAMED = (
        "[AUTOGEN] by pydependence resolver {resolver_name} **DO NOT EDIT** [AUTOGEN]"
    )

    @property
    def autogen_notice(self) -> str:
        if self.resolver_name is None:
            return self._AUTOGEN_NOTICE
        else:
            return self._AUTOGEN_NOTICE_NAMED.format(
                resolver_name=repr(self.resolver_name)
            )

    def _get_debug_struct(self) -> "List[Tuple[str, List[str]]]":
        return [
            (req.requirement, [src.source_module for src in req.sources])
            for req in self.requirements
        ]

    def as_requirements_txt(
        self,
        notice: bool = True,
        sources: bool = True,
        sources_compact: bool = False,
        sources_roots: bool = False,
        sources_annotations: bool = True,
        indent_size: int = 4,
    ) -> str:
        lines = []
        if notice:
            lines.append(self.autogen_notice)
        for req in self.requirements:
            # add requirement
            lines.append(f"{req.requirement}")
            # add annotations
            lines[
                -1
            ] += f"{req.get_annotations_string(enabled=sources_annotations, comment=True)}"
            # add compact sources
            if sources:
                if sources_compact:
                    lines[-1] += f" # {req.get_sources_string(roots=sources_roots)}"
                else:
                    for src_info in req.get_source_info(roots=sources_roots):
                        lines.append(f"{' '*indent_size*1}# {src_info.anno_str}")
        if self.requirements or notice:
            lines.append("")
        return "\n".join(lines)

    def as_toml_array(
        self,
        notice: bool = True,
        sources: bool = True,
        sources_compact: bool = False,
        sources_roots: bool = False,
        sources_annotations: bool = True,
        indent_size: int = 4,
    ):
        import tomlkit
        import tomlkit.container
        import tomlkit.items

        # create table
        array = tomlkit.array().multiline(True)
        if notice:
            array.add_line(
                indent=" " * (indent_size * 1),
                comment=self.autogen_notice,
            )
        for req in self.requirements:
            comment = ""
            comment += req.get_annotations_string(
                enabled=sources_annotations,
                comment=False,
            )
            comment += req.get_sources_string(
                enabled=sources and sources_compact,
                roots=sources_roots,
            )
            # add requirement & compact sources
            array.add_line(
                req.requirement,
                indent=" " * (indent_size * 1),
                comment=comment,
            )
            # add extended sources
            for src_info in req.get_source_info(
                enabled=sources and not sources_compact,
                roots=sources_roots,
            ):
                # Add line has a bug where it doesn't add the correct indentation before the comment
                # - so we instead add padding after the `#`
                # `array.add_line(indent=f"{' ' * (indent_size * 1)}, comment=f"{src_info.anno_str}")`
                array.add_line(
                    indent="", comment=f"{' ' * (indent_size * 1)}{src_info.anno_str}"
                )
                # NOTE: While this does not properly handle commas between lines ...
                # array.append(tomlkit.items.Comment(tomlkit.container.Trivia(
                #     indent=" " * (indent_size * 1),
                #     comment_ws="",
                #     comment=f"# {src_info.anno_str}\n",
                #     trail="",
                # )))

        if self.requirements or notice:
            array.add_line(indent="")
        # done!
        return array


# ========================================================================= #
# END                                                                       #
# ========================================================================= #



# ============================================================================== #
# MIT License                                                                    #
#                                                                                #
# Copyright (c) 2024 Nathan Juraj Michlo                                         #
#                                                                                #
# Permission is hereby granted, free of charge, to any person obtaining a copy   #
# of this software and associated documentation files (the "Software"), to deal  #
# in the Software without restriction, including without limitation the rights   #
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell      #
# copies of the Software, and to permit persons to whom the Software is          #
# furnished to do so, subject to the following conditions:                       #
#                                                                                #
# The above copyright notice and this permission notice shall be included in all #
# copies or substantial portions of the Software.                                #
#                                                                                #
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR     #
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,       #
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE    #
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER         #
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,  #
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE  #
# SOFTWARE.                                                                      #
# ============================================================================== #

import pkgutil
import warnings
from importlib.machinery import FileFinder
from pathlib import Path
from typing import Iterator, NamedTuple

from pydependence._core.utils import assert_valid_import_name, assert_valid_tag

# ========================================================================= #
# MODULE INFO                                                               #
# ========================================================================= #


class ModuleMetadata(NamedTuple):
    path: Path
    name: str
    ispkg: bool

    # tag e.g. if package `yolov5` package loaded this, the `utils` module is not unique...
    tag: str

    @property
    def root_name(self) -> str:
        return self.name.split(".")[0]

    @property
    def tagged_name(self):
        return f"{self.tag}:{self.name}"

    @property
    def is_name_valid(self):
        try:
            assert_valid_import_name(self.name)
            return True
        except Exception:
            return False

    @property
    def pkgutil_module_info(self):
        if not self.path.is_absolute():
            raise ValueError(f"Path must be absolute, got: {self.path}")
        return pkgutil.ModuleInfo(
            module_finder=FileFinder(path=str(self.path)),
            name=self.name,
            ispkg=self.ispkg,
        )

    @classmethod
    def from_root_and_subpath(
        cls,
        root: Path,
        subpath: Path,
        tag: str,
    ) -> "ModuleMetadata":
        if not root.is_absolute():
            raise ValueError(f"Root path must be absolute, got: {root}")
        if not subpath.is_absolute():
            subpath = root / subpath
        if not subpath.name.endswith(".py"):
            raise ValueError(f"Subpath must be a python file, got: {subpath}")
        if not subpath.is_file():
            raise FileNotFoundError(f"Subpath must be an existing file, got: {subpath}")
        tag = assert_valid_tag(tag)
        rel = subpath.relative_to(root)
        if rel.name == "__init__.py":
            return ModuleMetadata(
                path=subpath,
                name=".".join(rel.parts[:-1]),
                ispkg=True,
                tag=tag,
            )
        else:
            return ModuleMetadata(
                path=subpath,
                name=".".join(rel.parts)[: -len(".py")],
                ispkg=False,
                tag=tag,
            )

    @classmethod
    def yield_search_path_modules(
        cls, search_path: Path, *, tag: str, valid_only: bool = True
    ) -> "Iterator[ModuleMetadata]":
        def _visit(p: Path):
            if p.is_dir():
                yield from p.glob("**/*.py")
            else:
                raise ValueError(f"Invalid path: {p}")

        for p in _visit(search_path):
            m = cls.from_root_and_subpath(search_path, subpath=p, tag=tag)
            if valid_only and (not m.is_name_valid):
                warnings.warn(
                    f"Invalid module name: {m.name}, cannot be imported or resolved, skipping: {m.path}"
                )
                continue
            yield m
        # Only one level deep & does not work if __init__.py is not present.
        # yield from pkgutil.iter_modules(path=[str(search_path)], prefix='')

    @classmethod
    def yield_package_modules(
        cls, package_path: Path, *, tag: str, valid_only: bool = True
    ) -> "Iterator[ModuleMetadata]":
        def _visit(p: Path):
            if p.is_file():
                yield p
            elif p.is_dir():
                yield from p.glob("**/*.py")
            else:
                raise ValueError(f"Invalid path: {p}")

        for p in _visit(package_path):
            m = cls.from_root_and_subpath(package_path.parent, subpath=p, tag=tag)
            if valid_only and (not m.is_name_valid):
                warnings.warn(
                    f"Invalid module name: {m.name}, cannot be imported or resolved, skipping: {m.path}"
                )
                continue
            yield m
        # Only one level deep & does not work if __init__.py is not present.
        # yield from pkgutil.iter_modules(path=[str(package_path)], prefix=f"{package_path.name}.")


# ========================================================================= #
# END                                                                       #
# ========================================================================= #


__all__ = ("ModuleMetadata",)



# ============================================================================== #
# MIT License                                                                    #
#                                                                                #
# Copyright (c) 2024 Nathan Juraj Michlo                                         #
#                                                                                #
# Permission is hereby granted, free of charge, to any person obtaining a copy   #
# of this software and associated documentation files (the "Software"), to deal  #
# in the Software without restriction, including without limitation the rights   #
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell      #
# copies of the Software, and to permit persons to whom the Software is          #
# furnished to do so, subject to the following conditions:                       #
#                                                                                #
# The above copyright notice and this permission notice shall be included in all #
# copies or substantial portions of the Software.                                #
#                                                                                #
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR     #
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,       #
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE    #
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER         #
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,  #
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE  #
# SOFTWARE.                                                                      #
# ============================================================================== #


from pathlib import Path

import pytest

import pydependence._colors as C
from pydependence._core.utils import (
    apply_root_to_path_str,
    assert_valid_import_name,
    assert_valid_module_path,
    assert_valid_tag,
)


def test_colors():
    for attr in [
        "RST",
        # dark colors
        "GRY",
        "lRED",
        "lGRN",
        "lYLW",
        "lBLU",
        "lMGT",
        "lCYN",
        "WHT",
        # light colors
        "BLK",
        "RED",
        "GRN",
        "YLW",
        "BLU",
        "MGT",
        "CYN",
        "lGRY",
    ]:
        assert hasattr(C, attr)


def test_assert_valid_tag():
    # does not normalize tag
    assert assert_valid_tag("valid_tag") == "valid_tag"
    assert assert_valid_tag("valid-tag") == "valid-tag"
    with pytest.raises(ValueError):
        assert_valid_tag("")


def test_assert_valid_module_path():
    assert assert_valid_module_path(Path(__file__)) == Path(__file__).resolve()
    with pytest.raises(ValueError):
        assert_valid_module_path("relative/path")
    with pytest.raises(FileNotFoundError):
        assert_valid_module_path("/path/does/not/exist")
    with pytest.raises(RuntimeError):
        assert_valid_module_path(Path.home())  # assuming home directory is not a file


def test_assert_valid_import_name():
    assert assert_valid_import_name("valid.import.name") == "valid.import.name"
    with pytest.raises(NameError):
        assert_valid_import_name("")
    with pytest.raises(NameError):
        assert_valid_import_name("invalid.import.name.")


def test_apply_root_to_path_str():
    root = str(Path.home())
    assert apply_root_to_path_str(root, "relative/path") == str(
        (Path(root) / "relative/path").resolve()
    )
    with pytest.raises(ValueError):
        apply_root_to_path_str("relative/path", "another/relative/path")
    assert apply_root_to_path_str(root, str(Path.home())) == str(Path.home().resolve())



# ============================================================================== #
# MIT License                                                                    #
#                                                                                #
# Copyright (c) 2024 Nathan Juraj Michlo                                         #
#                                                                                #
# Permission is hereby granted, free of charge, to any person obtaining a copy   #
# of this software and associated documentation files (the "Software"), to deal  #
# in the Software without restriction, including without limitation the rights   #
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell      #
# copies of the Software, and to permit persons to whom the Software is          #
# furnished to do so, subject to the following conditions:                       #
#                                                                                #
# The above copyright notice and this permission notice shall be included in all #
# copies or substantial portions of the Software.                                #
#                                                                                #
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR     #
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,       #
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE    #
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER         #
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,  #
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE  #
# SOFTWARE.                                                                      #
# ============================================================================== #

from pathlib import Path
from typing import List, Union

# ========================================================================= #
# AST IMPORT PARSER                                                         #
# ========================================================================= #


def assert_valid_tag(tag: str) -> str:
    if not tag:
        raise ValueError(f"Tag must not be empty: {tag}")
    if not tag.replace("-", "_").isidentifier():
        raise NameError(f"Tag must be a valid identifier: {tag}")
    return tag


def assert_valid_module_path(path: "Union[Path, str]") -> Path:
    path = Path(path)
    if not path.is_absolute():
        raise ValueError(f"Path must be absolute: {path}")
    if not path.exists():
        raise FileNotFoundError(f"File does not exist: {path}")
    if not path.is_file():
        raise RuntimeError(f"Path is not a file: {path}")
    return path


def assert_valid_import_name(import_: str) -> str:
    parts = import_.split(".")
    if not parts:
        raise ValueError(
            f"import path must have at least one part for: {repr(import_)}"
        )
    for part in parts:
        if not part.isidentifier():
            raise NameError(
                f"import part: {repr(part)} is not a valid identifier, obtained from: {repr(import_)}"
            )
    return import_


# ========================================================================= #
# PATH HELPER                                                               #
# ========================================================================= #


def apply_root_to_path_str(root: "Union[str, Path]", path: "Union[str, Path]") -> str:
    root = Path(root)
    path = Path(path)
    if not root.is_absolute():
        raise ValueError(f"root must be an absolute path, got: {root}")
    if path.is_absolute():
        merged = path
    else:
        merged = root / path
    return str(merged.resolve())


# ========================================================================= #
# LOAD                                                                      #
# ========================================================================= #


def load_toml_document(
    path: "Union[str, Path]",
) -> "tomlkit.toml_document.TOMLDocument":
    import tomlkit
    import tomlkit.items
    import tomlkit.toml_document

    path = Path(path)
    # if not path.name.endswith(".toml"):
    #     raise ValueError(f"path is not a .toml file: {path}")
    if not path.is_file():
        raise FileNotFoundError(f"path is not a file: {path}")
    with open(path) as fp:
        toml = tomlkit.load(fp)
        assert isinstance(
            toml, tomlkit.toml_document.TOMLDocument
        ), f"got {type(toml)}, not TOMLDocument"
    return toml


# ========================================================================= #
# WRITE                                                                     #
# ========================================================================= #


def txt_file_dump(
    *,
    file: "Union[str, Path]",
    contents: "str",
):
    # write
    with open(file, "w") as fp:
        fp.write(contents)
        if not contents.endswith("\n"):
            fp.write("\n")


def toml_file_replace_array(
    *,
    file: "Union[str, Path]",
    keys: "List[str]",
    array: "tomlkit.items.Array",
):
    import tomlkit
    import tomlkit.items
    import tomlkit.toml_document

    # TODO: this needs multiple loads and writes if we are modifying multiple arrays
    #       which is not ideal... but it is a simple solution for now.

    assert isinstance(
        array, tomlkit.items.Array
    ), f"array must be a tomlkit Array, got: {type(array)}"

    # load file
    file = Path(file)
    assert file.is_absolute(), f"file must be an absolute path, got: {file}"
    toml = load_toml_document(file)

    # split parent keys from array key
    (*parent_keys, array_key) = keys

    # add parent sections if missing
    parent = toml
    for i, k in enumerate(parent_keys):
        section = parent.setdefault(k, {})
        assert isinstance(section, tomlkit.items.Table)
        parent = section

    # set array
    if array_key in parent:
        old_array = parent[array_key]
        assert isinstance(old_array, tomlkit.items.Array)
    parent[array_key] = array

    # write
    with open(file, "w") as fp:
        tomlkit.dump(toml, fp)


# ========================================================================= #
# END                                                                       #
# ========================================================================= #


__all__ = (
    "assert_valid_module_path",
    "assert_valid_import_name",
    "apply_root_to_path_str",
    "load_toml_document",
)



# ============================================================================== #
# MIT License                                                                    #
#                                                                                #
# Copyright (c) 2024 Nathan Juraj Michlo                                         #
#                                                                                #
# Permission is hereby granted, free of charge, to any person obtaining a copy   #
# of this software and associated documentation files (the "Software"), to deal  #
# in the Software without restriction, including without limitation the rights   #
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell      #
# copies of the Software, and to permit persons to whom the Software is          #
# furnished to do so, subject to the following conditions:                       #
#                                                                                #
# The above copyright notice and this permission notice shall be included in all #
# copies or substantial portions of the Software.                                #
#                                                                                #
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR     #
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,       #
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE    #
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER         #
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,  #
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE  #
# SOFTWARE.                                                                      #
# ============================================================================== #


import ast
import dataclasses
import sys
import warnings
from collections import Counter, defaultdict
from enum import Enum
from typing import DefaultDict, Dict, List, Literal, NamedTuple, Optional, Tuple

from pydependence._core.module_data import ModuleMetadata
from pydependence._core.utils import assert_valid_import_name, assert_valid_module_path

# ========================================================================= #
# Polyfill                                                                  #
# ========================================================================= #


def ast_unparse(node: ast.AST) -> str:
    # if not python 3.8 then call ast.unparse, otherwise polyfill
    if hasattr(ast, "unparse"):
        return ast.unparse(node)
    else:
        warnings.warn(
            f"Current version of python: {sys.version_info} does not support `ast.unparse`"
        )
        return str(node)


# ========================================================================= #
# Allowed Nodes Leading To Imports                                          #
# ========================================================================= #

# PYTHON VERSION: 3.10
# every possible way to reach an import statement, or because we also choose to
# support plugins, all expressions should be reachable. We skip the AST nodes
# that do not matter in reaching these.
# - https://docs.python.org/3/library/ast.html
_DISALLOWED_IMPORT_STATEMENT_NODES = {
    #          -- the following expression can appear in assignment context
    #          | Name(identifier id, expr_context ctx)
    "Name",
    #     | Global(identifier* names)
    #     | Nonlocal(identifier* names)
    #     | Pass | Break | Continue
    #     | Constant(constant value, string? kind)
    "Global",
    "Nonlocal",
    "Pass",
    "Break",
    "Continue",
    "Constant",
    #     alias = (identifier name, identifier? asname)
    #              attributes (int lineno, int col_offset, int? end_lineno, int? end_col_offset)
    "alias"
    #     attributes (int lineno, int col_offset, int end_lineno, int end_col_offset)
    "attributes",
    #     type_ignore = TypeIgnore(int lineno, string tag)
    "TypeIgnore",
    # ================================================================================ #
    #     expr_context = Load | Store | Del
    "Load",
    "Store",
    "Del",
    #     boolop = And | Or
    "And",
    "Or",
    #     operator = Add | Sub | Mult | MatMult | Div | Mod | Pow | LShift
    #                  | RShift | BitOr | BitXor | BitAnd | FloorDiv
    "Add",
    "Sub",
    "Mult",
    "MatMult",
    "Div",
    "Mod",
    "Pow",
    "LShift",
    "RShift",
    "BitOr",
    "BitXor",
    "BitAnd",
    "FloorDiv",
    #     unaryop = Invert | Not | UAdd | USub
    "Invert",
    "Not",
    "UAdd",
    "USub",
    #     cmpop = Eq | NotEq | Lt | LtE | Gt | GtE | Is | IsNot | In | NotIn
    "Eq",
    "NotEq",
    "Lt",
    "LtE",
    "Gt",
    "GtE",
    "Is",
    "IsNot",
    "In",
    "NotIn",
}

# these nodes are considered to cause lazy evaluation of imports. Side effects
# on a module level for example should be considered unsupported behavior and should
# never be considered for triggering an import.
_IS_INDIRECT_NODE = {
    "FunctionDef",
    "AsyncFunctionDef",
}


# lazy import callables
_LAZY_IMPORT_CALLABLES = {
    "lazy_import",
}

# lazy attribute callables
_LAZY_ATTRIBUTE_CALLABLES = {
    "lazy_callable",
    "lazy_inheritable",
}

# lazy
_LAZY_CALLABLES = {*_LAZY_IMPORT_CALLABLES, *_LAZY_ATTRIBUTE_CALLABLES}


# ========================================================================= #
# AST IMPORT PARSER                                                         #
# ========================================================================= #


class ImportSourceEnum(str, Enum):
    import_ = "import_"
    import_from = "import_from"
    lazy_plugin = "lazy_plugin"
    # type_check = 'type_check'  # TODO


@dataclasses.dataclass
class BasicImportInfo:
    # target
    target: str
    source_name: str
    is_lazy: bool

    @property
    def root_target(self) -> str:
        return self.target.split(".")[0]


class ManualSource:
    def __init__(self, orig_name: str):
        self.orig_name = orig_name

    def __str__(self):
        return f"<manual: {self.orig_name}>"

    def __repr__(self):
        return f"{self.__class__.__name__}({self.orig_name})"

    def __eq__(self, other):
        if isinstance(other, str):
            return str(self) == other
        elif isinstance(other, ManualSource):
            return self.orig_name == other.orig_name
        return False

    def __lt__(self, other):
        if isinstance(other, str):
            return str(self) < other
        elif isinstance(other, ManualSource):
            return self.orig_name < other.orig_name
        else:
            raise TypeError(f"Invalid types: {type(self)} < {type(other)}")

    def __hash__(self):
        return hash(str(self))


@dataclasses.dataclass
class ManualImportInfo(BasicImportInfo):
    source_name: ManualSource
    is_lazy: Literal[False] = False

    @classmethod
    def from_target(cls, target: str) -> "ManualImportInfo":
        return cls(target=target, source_name=ManualSource(target))


@dataclasses.dataclass
class LocImportInfo(BasicImportInfo):
    # source, e.g. import statement or type check or lazy plugin
    source_name: str
    source_module_info: ModuleMetadata
    source_type: ImportSourceEnum
    # debug
    lineno: int
    col_offset: int
    stack_type_names: Tuple[str, ...]
    # relative import
    is_relative: bool

    @property
    def tagged_target(self) -> str:
        return f"{self.source_module_info.tag}:{self.target}"

    @property
    def tagged_name_and_target(self) -> str:
        return f"{self.source_module_info.tagged_name}:{self.target}"


class _AstImportsCollector(ast.NodeVisitor):

    def __init__(self, module_info: ModuleMetadata):
        self._module_info: ModuleMetadata = module_info
        self._imports: "DefaultDict[str, List[LocImportInfo]]" = defaultdict(list)
        self._stack_is_lazy: "List[bool]" = [False]
        self._stack_ast_kind: "List[str]" = []
        self._counter = Counter()

    # ~=~=~ WARN ~=~=~ #

    def _node_warn(self, node: ast.AST, message: str):
        warnings.warn_explicit(
            message=f"`{ast_unparse(node)}`: {message}",
            category=SyntaxWarning,
            filename=str(self._module_info.path),
            lineno=node.lineno,
        )

    # ~=~=~ STACK ~=~=~ #

    def _push_current_import(
        self,
        node: ast.AST,
        target: str,
        source_type: ImportSourceEnum,
        is_lazy: "Optional[bool]" = None,
        is_relative: bool = False,
    ):
        import_ = LocImportInfo(
            source_name=self._module_info.name,
            source_module_info=self._module_info,
            target=target,
            is_lazy=self._stack_is_lazy[-1] if (is_lazy is None) else is_lazy,
            lineno=node.lineno,
            col_offset=node.col_offset,
            source_type=source_type,
            stack_type_names=tuple(self._stack_ast_kind),
            is_relative=is_relative,
        )
        self._imports[target].append(import_)

    # ~=~=~ VISIT ~=~=~ #

    def visit(self, node):
        kind = node.__class__.__name__
        if kind in _DISALLOWED_IMPORT_STATEMENT_NODES:
            return
        # push - basic interpreter
        is_lazy = self._stack_is_lazy[-1] or (kind in _IS_INDIRECT_NODE)
        self._stack_ast_kind.append(kind)
        self._stack_is_lazy.append(is_lazy)
        # continue recursion
        try:
            getattr(self, "visit_" + kind, self.generic_visit)(node)
        finally:
            # pop
            self._stack_is_lazy.pop()
            self._stack_ast_kind.pop()

    def generic_visit_lazy(self, node):
        self._stack_is_lazy.append(True)
        try:
            self.generic_visit(node)
        finally:
            self._stack_is_lazy.pop()

    # >>> VISIT NODES <<< #

    def visit_FunctionDef(self, node):
        return self.generic_visit_lazy(node)

    def visit_AsyncFunctionDef(self, node):
        return self.generic_visit_lazy(node)

    def visit_Import(self, node: ast.Import):
        # eg. import pkg.submodule
        for alias in node.names:
            self._push_current_import(
                node=node,
                target=alias.name,
                source_type=ImportSourceEnum.import_,
            )

    def visit_ImportFrom(self, node: ast.ImportFrom):
        assert node.level in (0, 1)  # node.names: from * import name, ...
        # eg: from . import ?
        # eg: from .submodule import ?
        # eg: from pkg.submodule import ?
        is_relative = node.level != 0
        if is_relative:
            _parts = self._module_info.name.split(".")
            if not self._module_info.ispkg:
                _parts.pop()
            _parts.append(node.module)
            target = ".".join(_parts)
            assert_valid_import_name(target)
        else:
            target = node.module
        self._push_current_import(
            node=node,
            target=target,
            source_type=ImportSourceEnum.import_from,
            is_relative=is_relative,
        )

    # >>> CUSTOM LAZY IMPORT LIBRARY

    def visit_If(self, node: ast.If):
        """
        check name is `TYPE_CHECKING` or attr is `typing.TYPE_CHECKING`:
        - WE DON'T SUPPORT ANY OTHER VARIATIONS
        - WE ASSUME THE VARIABLE IS NEVER RENAMED
        """
        # check if this is a type checking block
        is_type_checking = False
        if isinstance(node.test, ast.Attribute):
            if (
                isinstance(node.test.value, ast.Name)
                and node.test.value.id == "typing"
                and node.test.attr == "TYPE_CHECKING"
            ):
                is_type_checking = True
        elif isinstance(node.test, ast.Name):
            if node.test.id == "TYPE_CHECKING":
                is_type_checking = True
        # recurse
        if not is_type_checking:
            return self.generic_visit(node)
        else:
            return self.generic_visit_lazy(node)

    def visit_Call(self, node: ast.Call):
        """
        we don't implement an interpreter, we only handle lazy imports with an
        exact function name and a single string argument. These functions should
        be defined by the user. We do not provide any default functions.
          e.g. `lazy_import("os.path")`
        we don't support attribute access or any deviation from the above
          e.g. `util.lazy_import("os.path.join")`

        - the function names must be one of:
          * `lazy_import("os.path")`
          * `lazy_callable("pathlib.Attr")`  # final .Path is excluded, import is `pathlib`
          * `lazy_inheritable("pathlib.Attr")`  # final .Path is excluded, import is `pathlib`
          - WE DON'T SUPPORT ANY OTHER VARIATIONS
          - WE ASSUME THE VARIABLE IS NEVER RENAMED
        """
        # - check the call is directly on a name e.g. `lazy_import(...)` and not `util.lazy_import(...)` or `util['lazy_import'](...)`
        if not isinstance(node.func, ast.Name):
            return
        # - check the function name is one of the lazy import functions
        name = node.func.id
        if name not in _LAZY_CALLABLES:
            return
        # - make sure no keyword arguments are used, these invalidate the import.
        if node.keywords:
            self._node_warn(node, f"should not have keyword arguments.")
            return
        # - make sure that the function is called with a single string argument
        if not len(node.args) == 1:
            self._node_warn(
                node, f"called with {len(node.args)} arguments, expected: 1"
            )
            return
        [arg] = node.args
        # - make sure that the argument is a string
        if not isinstance(arg, ast.Constant) or not isinstance(arg.value, str):
            self._node_warn(
                node, f"called with non-string argument: `{ast_unparse(arg)}`"
            )
            return
        # - validate the import string
        import_ = arg.value
        try:
            assert_valid_import_name(import_)
        except Exception as e:
            self._node_warn(node, f"called with invalid import path: {e}")
            return
        # - check if the import path includes an attribute and strip it
        if name in _LAZY_ATTRIBUTE_CALLABLES:
            _parts = import_.rsplit(".", maxsplit=1)
            if len(_parts) < 2:
                self._node_warn(
                    node, f"called with invalid import path to an attribute: {import_}"
                )
                return
            import_ = _parts[0]
        # - add the import
        self._push_current_import(
            node=node,
            target=import_,
            source_type=ImportSourceEnum.lazy_plugin,
            is_lazy=True,
        )

    # >>> PRETTY PRINT <<< #

    @classmethod
    def _ast_to_dict(cls, n):
        if isinstance(n, ast.AST):
            return {f: cls._ast_to_dict(getattr(n, f)) for f in n._fields}
        elif isinstance(n, list):
            return [cls._ast_to_dict(i) for i in n]
        else:
            return n

    @classmethod
    def load_imports_from_module_info(
        cls, module_info: ModuleMetadata, *, debug: bool = False
    ) -> "Dict[str, List[LocImportInfo]]":
        # load the file & parse
        path = assert_valid_module_path(module_info.path)
        name = assert_valid_import_name(module_info.name)
        with open(path) as fp:
            _dat = fp.read()
            _ast = ast.parse(_dat)
        # collect imports
        _parser = _AstImportsCollector(module_info=module_info)
        _parser.visit(_ast)
        # debug
        if debug:
            total = sum(_parser._counter.values())
            top = _parser._counter.most_common(5)
            print(
                f"Visited {total} nodes, top 5: {top} for module: {repr(name)} file: {module_info.path}"
            )
        # done!
        return _parser._imports


def load_imports_from_module_info(
    module_info: ModuleMetadata,
) -> "Dict[str, List[LocImportInfo]]":
    return _AstImportsCollector.load_imports_from_module_info(module_info)


# ========================================================================= #
# END                                                                       #
# ========================================================================= #


__all__ = (
    "load_imports_from_module_info",
    "LocImportInfo",
    "ImportSourceEnum",
)


# PYTHON VERSION: 3.10
# {
#         #     mod = Module(stmt* body, type_ignore* type_ignores)
#         #         | Interactive(stmt* body)
#         #         | Expression(expr body)
#         #         | FunctionType(expr* argtypes, expr returns)
#         "Module",
#         "Interactive",
#         "Expression",
#         #     stmt = FunctionDef(identifier name, arguments args,
#         #                        stmt* body, expr* decorator_list, expr? returns,
#         #                        string? type_comment)
#         #           | AsyncFunctionDef(identifier name, arguments args,
#         #                              stmt* body, expr* decorator_list, expr? returns,
#         #                              string? type_comment)
#         #
#         "FunctionDef",
#         "AsyncFunctionDef",
#         #           | ClassDef(identifier name,
#         #              expr* bases,
#         #              keyword* keywords,
#         #              stmt* body,
#         #              expr* decorator_list)
#         #           | Return(expr? value)
#         "ClassDef",
#         "Return",
#         #           | Delete(expr* targets)
#         #           | Assign(expr* targets, expr value, string? type_comment)
#         #           | AugAssign(expr target, operator op, expr value)
#         "Delete",
#         "Assign",
#         "AugAssign",
#         #           -- 'simple' indicates that we annotate simple name without parens
#         #           | AnnAssign(expr target, expr annotation, expr? value, int simple)
#         "AnnAssign",
#         #           -- use 'orelse' because else is a keyword in target languages
#         #           | For(expr target, expr iter, stmt* body, stmt* orelse, string? type_comment)
#         #           | AsyncFor(expr target, expr iter, stmt* body, stmt* orelse, string? type_comment)
#         #           | While(expr test, stmt* body, stmt* orelse)
#         #           | If(expr test, stmt* body, stmt* orelse)
#         #           | With(withitem* items, stmt* body, string? type_comment)
#         #           | AsyncWith(withitem* items, stmt* body, string? type_comment)
#         "For",
#         "AsyncFor",
#         "While",
#         "If",
#         "With",
#         "AsyncWith",
#         #           | Match(expr subject, match_case* cases)  # stmt
#         "Match",
#         #           | Raise(expr? exc, expr? cause)
#         #           | Try(stmt* body, excepthandler* handlers, stmt* orelse, stmt* finalbody)
#         #           | TryStar(stmt* body, excepthandler* handlers, stmt* orelse, stmt* finalbody)
#         #           | Assert(expr test, expr? msg)
#         "Raise",
#         "Try",
#         "TryStar",
#         "Assert",
#         #           | Import(alias* names)
#         #           | ImportFrom(identifier? module, alias* names, int? level)
#         "Import",
#         "ImportFrom",
#         #           | Global(identifier* names)
#         #           | Nonlocal(identifier* names)
#         #           | Expr(expr value)
#         #           | Pass | Break | Continue
#         #
#         #           -- col_offset is the byte offset in the utf8 string the parser uses
#         #           attributes (int lineno, int col_offset, int? end_lineno, int? end_col_offset)
#         "Global",
#         "Nonlocal",
#         "Expr",
#         "Pass",
#         "Break",
#         "Continue",
#         #           -- BoolOp() can use left & right?
#         #     expr = BoolOp(boolop op, expr* values)
#         #          | NamedExpr(expr target, expr value)
#         #          | BinOp(expr left, operator op, expr right)
#         #          | UnaryOp(unaryop op, expr operand)
#         "BoolOp",
#         "NamedExpr",
#         "BinOp",
#         "UnaryOp",
#         #          | Lambda(arguments args, expr body)
#         #          | IfExp(expr test, expr body, expr orelse)
#         #          | Dict(expr* keys, expr* values)
#         #          | Set(expr* elts)
#         #          | ListComp(expr elt, comprehension* generators)
#         #          | SetComp(expr elt, comprehension* generators)
#         #          | DictComp(expr key, expr value, comprehension* generators)
#         #          | GeneratorExp(expr elt, comprehension* generators)
#         "Lambda",
#         "IfExp",
#         "Dict",
#         "Set",
#         "ListComp",
#         "SetComp",
#         "DictComp",
#         "GeneratorExp",
#         #          -- the grammar constrains where yield expressions can occur
#         #          | Await(expr value)
#         #          | Yield(expr? value)
#         #          | YieldFrom(expr value)
#         "Await",
#         "Yield",
#         "YieldFrom",
#         #          -- need sequences for compare to distinguish between
#         #          -- x < 4 < 3 and (x < 4) < 3
#         #          | Compare(expr left, cmpop* ops, expr* comparators)
#         #          | Call(expr func, expr* args, keyword* keywords)
#         #          | FormattedValue(expr value, int conversion, expr? format_spec)
#         #          | JoinedStr(expr* values)
#         #          | Constant(constant value, string? kind)
#         "Compare",
#         "Call",
#         "FormattedValue",
#         "JoinedStr",
#         "Constant",
#         #          -- the following expression can appear in assignment context
#         #          | Attribute(expr value, identifier attr, expr_context ctx)
#         #          | Subscript(expr value, expr slice, expr_context ctx)
#         #          | Starred(expr value, expr_context ctx)
#         #          | Name(identifier id, expr_context ctx)
#         #          | List(expr* elts, expr_context ctx)
#         #          | Tuple(expr* elts, expr_context ctx)
#         "Attribute",
#         "Subscript",
#         "Starred",
#         "Name",
#         "List",
#         "Tuple",
#         #          -- can appear only in Subscript
#         #          | Slice(expr? lower, expr? upper, expr? step)
#         "Slice",
#         #           -- col_offset is the byte offset in the utf8 string the parser uses
#         #           attributes (int lineno, int col_offset, int? end_lineno, int? end_col_offset)
#         #
#         #     expr_context = Load | Store | Del
#         "Load",
#         "Store",
#         "Del",
#         #     boolop = And | Or
#         "And",
#         "Or",
#         #     operator = Add | Sub | Mult | MatMult | Div | Mod | Pow | LShift
#         #                  | RShift | BitOr | BitXor | BitAnd | FloorDiv
#         "Add",
#         "Sub",
#         "Mult",
#         "MatMult",
#         "Div",
#         "Mod",
#         "Pow",
#         "LShift",
#         "RShift",
#         "BitOr",
#         "BitXor",
#         "BitAnd",
#         "FloorDiv",
#         #     unaryop = Invert | Not | UAdd | USub
#         "Invert",
#         "Not",
#         "UAdd",
#         "USub",
#         #     cmpop = Eq | NotEq | Lt | LtE | Gt | GtE | Is | IsNot | In | NotIn
#         "Eq",
#         "NotEq",
#         "Lt",
#         "LtE",
#         "Gt",
#         "GtE",
#         "Is",
#         "IsNot",
#         "In",
#         "NotIn",

#         #     comprehension = (expr target, expr iter, expr* ifs, int is_async)
#         "comprehension",
#         #     excepthandler = ExceptHandler(expr? type, identifier? name, stmt* body)
#         #                     attributes (int lineno, int col_offset, int? end_lineno, int? end_col_offset)
#         "ExceptHandler",
#         #     arguments = (arg* posonlyargs, arg* args, arg? vararg, arg* kwonlyargs,
#         #                  expr* kw_defaults, arg? kwarg, expr* defaults)
#         "arguments",
#         #     arg = (identifier arg, expr? annotation, string? type_comment)
#         #            attributes (int lineno, int col_offset, int? end_lineno, int? end_col_offset)
#         "arg",
#         #     -- keyword arguments supplied to call (NULL identifier for **kwargs)
#         #     keyword = (identifier? arg, expr value)
#         #                attributes (int lineno, int col_offset, int? end_lineno, int? end_col_offset)
#         "keyword",
#         #     -- import name with optional 'as' alias.
#         #     alias = (identifier name, identifier? asname)
#         #              attributes (int lineno, int col_offset, int? end_lineno, int? end_col_offset)
#         "alias",
#         #     withitem = (expr context_expr, expr? optional_vars)
#         "withitem",
#         #     match_case = (pattern pattern, expr? guard, stmt* body)
#         "matchcase",
#         #     pattern = MatchValue(expr value)
#         #             | MatchSingleton(constant value)
#         #             | MatchSequence(pattern* patterns)
#         #             | MatchMapping(expr* keys, pattern* patterns, identifier? rest)
#         #             | MatchClass(expr cls, pattern* patterns, identifier* kwd_attrs, pattern* kwd_patterns)
#         "MatchValue",
#         "MatchSingleton",
#         "MatchSequence",
#         "MatchMapping",
#         "MatchClass",
#         #             | MatchStar(identifier? name)
#         #             -- The optional "rest" MatchMapping parameter handles capturing extra mapping keys
#         "MatchStar",
#         #             | MatchAs(pattern? pattern, identifier? name)
#         #             | MatchOr(pattern* patterns)
#         "MatchAs",
#         "MatchOr",
#         #              attributes (int lineno, int col_offset, int end_lineno, int end_col_offset)
#         "attributes",
#         #     type_ignore = TypeIgnore(int lineno, string tag)
#         "TypeIgnore",
#
#         # visit methods for deprecated nodes
#         "ExtSlice",
#         "Index",
#         "Suite",
#         "AugLoad",
#         "AugStore",
#         "Param",
#         "Num",
#         "Str",
#         "Bytes",
#         "NameConstant",
#         "Ellipsis",
#     }



import os
from sys import path

from foo.bar import asdf

from .package import sub_module


def test_func():
    import json
    import sys

    from asdf.fdsa import foo as bar


asdf = lazy_import("buzz")
fdsa = lazy_callable("bazz")
foo = lazy_inheritable("bar")
